// Imports
#include <Adafruit_PWMServoDriver.h>
#include <Servo.h>

int counter = 0;

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// SENSORS
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

// Induction sensor
#define inductPin 2
bool bone;

// Ultrasonic sensor
#define trigPin 4
#define echoPin 5
float maxDist = 30;
float minDist = 5;
float distance;

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// SERVOS
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

// Servo multiplexer declaration
Adafruit_PWMServoDriver board1 = Adafruit_PWMServoDriver(0x40);// called this way, it uses the default address 0x40   

// Servo modulation limits
#define SERVOMIN  125// this is the 'minimum' pulse length count (out of 4096)
#define SERVOMAX  625


const float a = 5.0; // Upper leg length in cm
const float b = 7.5; // Lower leg length in cm
const float d = 1000; // Time delay in between movements
const float R = 2.0;
const unsigned long T = 500;
unsigned long StartTime;

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// PUMP
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

// Pump control pin
#define relayPin 7 // Pin connected to the relay's IN pin

void setup() {

  // Initialize components
  initInduction();
  initPump();
  initUltraSound();
  board1.begin();
  board1.setPWMFreq(60); 
  initServos();


  // Initialize Serial port
  Serial.begin(9600);
  Serial.println("Initialization complete");

}

void loop() {
  WalkForward();
  distance = getDistance();
  Serial.println("Distance: " + String(distance) + "cm");
  if(distance>maxDist){
    WalkForward();
  }
  else if(distance<=maxDist && distance >minDist){
    Serial.println("Waiting for bone");
    LookUp();
    delay(5000);
    bone = readInduction();
    if(bone){
      Serial.println("Bone detected. Activating paw.");
      Paw();
    }
    else{
      Serial.println("No bone detected. Peeing.");
      LookDown();
      pee(1000);

    while(distance<=maxDist && distance >minDist){
      WalkBackward();
      distance = getDistance();
    }
    }
  }
}

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// HIGH LEVEL FUNCTIONS
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

void pee(int peeTime){// Activate water pump for peeTime miliseconds
  digitalWrite(relayPin, HIGH); // Turn the relay on (pump on)
  delay(peeTime); // Wait for 1 second
  digitalWrite(relayPin, LOW); // Turn the relay off (pump off)
  delay(5000); // Wait for 5 seconds
}

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// INDUCTION SENSOR CONTROL
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

void initInduction(){// Initialize inductance sensor's pin
  pinMode(inductPin, INPUT);
}

bool readInduction(){// Check if the sensor detects an object
  int sensorValue = digitalRead(inductPin);
  bool detected = false;
  if(sensorValue==HIGH){ 
      detected = false;
      Serial.println("No object");
  }
  else if(sensorValue==LOW){
    detected = true;
    Serial.println(String("Object detected"));
  }
  delay(500);
  return detected;
}

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// ULTRASOUND SENSORS CONTROL
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

void initUltraSound(){// Initialize ultrasonic sensor
  pinMode(trigPin, OUTPUT);
  pinMode(echoPin, INPUT);
}

float getDistance(){// Gets the distances read by the ultrasonic sensor

  float duration, distance;

  digitalWrite(trigPin,LOW);
  delayMicroseconds(2);
  digitalWrite(trigPin,HIGH);
  delayMicroseconds(10);
  digitalWrite(trigPin,LOW);

  duration = pulseIn(echoPin,HIGH);
  distance = (duration/2) /29.1;

  // Serial.println(String(distance));
  return distance;
}

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// SERVO CONTROL
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

int angleToPulse(int ang){//Gets angle in degree and returns the pulse width
  int pulse = map(ang,0, 180, SERVOMIN,SERVOMAX);  // map angle of 0 to 180 to Servo min and Servo max 
  // Serial.print("Angle: ");Serial.print(ang);
  // Serial.print(" pulse: ");Serial.println(pulse);
  return pulse;
}

void moveServo(int servo, int ang){//Moves servo at index servo to angle ang
  board1.setPWM(servo, 0, angleToPulse(ang) );
  delay(100);
}

void initServos(){
  board1.setPWM(1, 0, angleToPulse(90) );
  delay(d);
  board1.setPWM(3, 0, angleToPulse(90) );
  delay(d);
  board1.setPWM(5, 0, angleToPulse(90) );
  delay(d);
  board1.setPWM(7, 0, angleToPulse(90) );
  delay(d);
  // board1.setPWM(9, 0, angleToPulse(90) );
  // delay(d);
  // board1.setPWM(11, 0, angleToPulse(90) );
  // delay(d);
  // board1.setPWM(13, 0, angleToPulse(90) );
  // delay(d);
  // board1.setPWM(15, 0, angleToPulse(90) );
  // delay(d);
}

// Function to calculate and set the Front Left joint angles
void setFLLegPosition(float x, float y) {

  // Calculate the distance from the origin to the point
  float c = 10.0 - y;

  // Calculate the angle of the knee joint (theta2)
  float theta2 = acos( (sq(a) + sq(b) - sq(c)) / (2 * a * b) );
  
  // Calculate the angle of the hip joint (theta1)
  float theta1 = acos( (sq(a) + sq(c) - sq(b)) / (2 * a * c) ) + atan(x/c);
  
  // Convert radians to degrees
  float theta1Deg = theta1 * 180.0 / PI;
  float theta2Deg = theta2 * 180.0 / PI;

  // Start angle offsets remapping
  float LeftHipAngle = abs(theta1Deg + 45);
  float LeftKneeAngle = abs(theta2Deg - 180);

  float RightHipAngle = abs(180 - (theta1Deg + 45));
  float RightKneeAngle = abs(180 + (theta2Deg - 180));
  
  // Set the servo positions
  board1.setPWM(1, 0, angleToPulse(LeftHipAngle) );
  board1.setPWM(3, 0, angleToPulse(LeftKneeAngle) );

}

// Function to calculate and set the Front Right joint angles
void setFRLegPosition(float x, float y) {

  // Calculate the distance from the origin to the point
  float c = 10.0 - y;

  // Calculate the angle of the knee joint (theta2)
  float theta2 = acos( (sq(a) + sq(b) - sq(c)) / (2 * a * b) );
  
  // Calculate the angle of the hip joint (theta1)
  float theta1 = acos( (sq(a) + sq(c) - sq(b)) / (2 * a * c) ) + atan(x/c);
  
  // Convert radians to degrees
  float theta1Deg = theta1 * 180.0 / PI;
  float theta2Deg = theta2 * 180.0 / PI;

  // Start angle offsets remapping
  float LeftHipAngle = abs(theta1Deg + 45);
  float LeftKneeAngle = abs(theta2Deg - 180);

  float RightHipAngle = abs(180 - (theta1Deg + 45));
  float RightKneeAngle = abs(180 + (theta2Deg - 180));
  
  // Set the servo positions
  board1.setPWM(5, 0, angleToPulse(LeftHipAngle) );
  board1.setPWM(7, 0, angleToPulse(LeftKneeAngle) );

}


int CurrentLeg = 0;

void WalkForward(){

  Serial.println("Walking Forward");
  unsigned long CurrentTime = millis();
  unsigned long ElapsedTime = CurrentTime - StartTime;

  if (ElapsedTime > T) {
    StartTime = millis();
    ElapsedTime = 0;
    CurrentLeg = CurrentLeg + 1;
  }
  
  float t = (float(ElapsedTime) / float(T));
  float targetX, targetY;
  float theta = (PI + 0.1) * t * 2;

  if (t <= 0.5) {
    targetX = R * cos(theta); // Target x-coordinate in cm
    targetY = R * sin(theta); // Target y-coordinate in cm
  }

  else {
    float t2 = (t - 0.5) * 2;
    targetX = - R + (2 * R * t2);
    targetY = 0;
  }



  if (CurrentLeg == 0) {
    setFLLegPosition(targetX, targetY);
  }

  else if (CurrentLeg == 1) {
    setFRLegPosition(targetX, targetY);
  }

  // else if (CurrentLeg == 2) {
  //   setFRLegPosition(targetX, targetY);
  // }

  // else if (CurrentLeg == 3) {
  //   setBLLegPosition(targetX, targetY);
  // }

  else if (CurrentLeg > 1) {
    CurrentLeg = 0;
  }

}

void WalkBackward() {
  Serial.println("Walking Backward");
  unsigned long CurrentTime = millis();
  unsigned long ElapsedTime = CurrentTime - StartTime;

  if (ElapsedTime > T) {
    StartTime = millis();
    ElapsedTime = 0;
    CurrentLeg = CurrentLeg + 1;
  }
  
  float t = (float(ElapsedTime) / float(T));
  float targetX, targetY;
  float theta = (PI + 0.1) * t * 2;

  if (t <= 0.5) {
    targetX = - R * cos(theta); // Target x-coordinate in cm
    targetY = R * sin(theta); // Target y-coordinate in cm
  }

  else {
    float t2 = (t - 0.5) * 2;
    targetX = - (- R + (2 * R * t2));
    targetY = 0;
  }


  if (CurrentLeg == 0) {
    setFLLegPosition(targetX, targetY);
  }


  else if (CurrentLeg == 2) {
    setFRLegPosition(targetX, targetY);
  }


  else if (CurrentLeg > 3) {
    CurrentLeg = 0;
  }

}


void Paw() {
  Serial.println("Giving Paw");
  setFLLegPosition(0, 0);
  setFRLegPosition(-50, 0);

}

void LookUp() {
  Serial.println("Looking Up");
  setFLLegPosition(0, -5);
  setFRLegPosition(0, -5);
}

void LookDown() {
  Serial.println("Looking Down");
  setFLLegPosition(0, 5);
  setFRLegPosition(0, 5);
}

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// PUMP CONTROL
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

void initPump(){// Initializes water pump's pin
  pinMode(relayPin, OUTPUT); // Set the relay pin as an output
  digitalWrite(relayPin, LOW); // Ensure the relay is off at startup
}